#ifndef lint
static const char rcsid[] =
	"$Id: blimitd.c,v 1.4 2001/08/04 14:11:31 andrew Exp $";
#endif

#include <sys/stat.h>
#include <sys/types.h>

#include <err.h>
#include <pwd.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <sysexits.h>
#include <syslog.h>
#include <unistd.h>
#include <utmp.h>

#include "blimitd.h"
#include "config.h"
#include "idletime.h"
#include "utils.h"

unsigned int debug,
			 monitor_interval;
static volatile sig_atomic_t sig_no;

static void exit_on_signal(int local_sig_no);
static void usage(void);

int main(int argc, char *argv[]) {
	unsigned int idletime_on;
	int c;
	pid_t pid;
	char *name;
	FILE *utmp_fh;
	struct utmp utmp_record;
	struct passwd *pw;
	struct user_instance ui;

	/* set defaults */
	debug = 0;
	idletime_on = 1;
	monitor_interval = MONITOR_INTERVAL;
	sig_no = 0;

	/* grab command line arguments */
	while ((c = getopt(argc, argv, "dim:")) != -1) {
		switch (c) {
		case 'd':
			/* debug */
			++debug;
			break;
		case 'i':
			/* disable enforcement of idletime */
			idletime_on = 0;
			break;
		case 'm':
			/* monitor interval */
			if ((monitor_interval = atoi(optarg)) <= 0) {
				usage();
			}
			break;
		case '?':
		default:
			usage();
			break;
		}
	}
	argc -= optind;
	argv += optind;

	if (argc > 1) {
		usage();
	}

	/* background ourselves */
	if (! debug) {
		pid = fork();
		if (pid < 0) {
			/* error */
			err(EX_OSERR, "fork failed");
		} else if (pid > 0) {
			/* parent */
			exit(EX_OK);
		} else {
			/* child */
			(void)setsid();
			(void)chdir("/");
			(void)umask((mode_t)0);
			(void)close(STDIN_FILENO);
			(void)close(STDOUT_FILENO);
			(void)close(STDERR_FILENO);
		}
	}

	/* set up logging */
	openlog(SYSLOG_IDENT, LOG_PID | (debug ? LOG_PERROR : 0), SYSLOG_FACILITY);
	syslog(LOG_INFO, "starting up");

	/* handle pid file (borrow utmp_fh to save defining a new variable) */
	if ((utmp_fh = fopen(PID_FILE, "w")) != NULL) {
		fprintf(utmp_fh, "%ld", (long)getpid());
		fclose(utmp_fh);
	} else {
		syslog(LOG_ERR, "failed to open %s for writing: %m", PID_FILE);
	}

	/* attach signal handler */
	if ((signal(SIGINT, exit_on_signal) == SIG_ERR) ||
		(signal(SIGTERM, exit_on_signal) == SIG_ERR)) {
		syslog(LOG_ERR, "failed to install signal handler: %m");
	}

	/* actually do stuff */
	while (sig_no == 0) {
		/* get the logged in users */
		if ((utmp_fh = fopen(_PATH_UTMP, "r")) == NULL) {
			syslog(LOG_ERR, "failed to open %s for reading: %m", _PATH_UTMP);
			continue;
		}
		/* prepare for pw access */
		if (! setpassent(1)) {
			syslog(LOG_ERR, "setpassent: %m");
		}
		/* for each entry */
		while (fread(&utmp_record, sizeof(utmp_record), 1, utmp_fh)) {
			if (utmp_record.ut_name[0] == '\0') {
				/* this entry has been removed */
				continue;
			}
			/* get their name in a char *\0 */
			if ((name = strndup(utmp_record.ut_name, UT_NAMESIZE)) == NULL) {
				syslog(LOG_ERR, "strndup failed: %m");
				continue;
			}
			/* get their passwd entry */
			pw = getpwnam(name);

			/* now we (might) have the struct passwd we dont need name anymore
			 */
			free(name);

			/* make sure the user really exists */
			if (pw == NULL) {
				syslog(LOG_WARNING, "getpwnam for %s failed: %m", name);
				continue;
			}

			/* initialise ui for passing to enforcement functions */
			ui.pw = pw;
			if ((ui.tty = find_tty(utmp_record.ut_line, UT_LINESIZE)) == NULL) {
				syslog(LOG_ERR, "find_tty failed: %m");
				continue;
			}
			/* enforce limits */
			if (idletime_on) {
				enforce_idletime(&ui);
			}

			/* clean up ui for next instance */
			free(ui.tty);
		}

		if (ferror(utmp_fh)) {
			syslog(LOG_ERR, "error reading from %s: %m", _PATH_UTMP);
		}
		if (fclose(utmp_fh) != 0) {
			syslog(LOG_ERR, "error closing %s: %m", _PATH_UTMP);
		}

		/* finished with the password database till next time */
		endpwent();

		/* hang around until it's time to check again (a signal should
		 * interupt this) */
		sleep(monitor_interval);
	};
	
	syslog(LOG_INFO, "exiting on signal %d", (int)sig_no);
	return(EX_OK);
}

static void usage(void) {

	(void)fprintf(stderr, "usage: blimitd [-di] [-m time]\n");
	exit(EX_USAGE);
}

static void exit_on_signal(int local_sig_no) {

	sig_no = local_sig_no;
}
