#ifndef lint
static const char rcsid[] =
	"$Id: idletime.c,v 1.3 2001/08/04 14:11:31 andrew Exp $";
#endif

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/resource.h>

#include <login_cap.h>
#include <stddef.h>
#include <syslog.h>
#include <time.h>

#include "blimitd.h"
#include "config.h"
#include "idletime.h"
#include "kill.h"
#include "utils.h"

void enforce_idletime(struct user_instance *ui) {
	login_cap_t *lcap;
	rlim_t allowed_idletime;
	struct stat sb;
	unsigned int i;
	unsigned int warn_times[] = IDLE_WARN_TIMES;
	unsigned long long idletime;

	/* get the user's login_cap entry */
	if ((lcap = login_getpwclass(ui->pw)) == NULL) {
		syslog(LOG_ERR, "failed to get login class for user %s: %m", ui->pw->pw_name);
		return;
	}

	/* get their allowed idle time */
	allowed_idletime = login_getcaptime(lcap, "idletime", RLIM_INFINITY, RLIM_INFINITY);

	/* finished with the login.conf database now */
	login_close(lcap);

	/* check to see if the user is subject to this limit */
	if (allowed_idletime == RLIM_INFINITY) {
		/* they can idle as long as they want */
		if (debug) {
			syslog(LOG_DEBUG, "user %s is allowed to idle on %s", ui->pw->pw_name, ui->tty);
		}
		return;
	}

	/* find out how long the user has been idle for */
	if (stat(ui->tty, &sb) == -1) {
		/* failed to stat */
		syslog(LOG_ERR, "stat of %s failed: %m", ui->tty);
		return;
	}

	/* see if that is too long */
	if ((idletime = time(NULL) - sb.st_atime) > (unsigned long long)allowed_idletime) {
		/* time to kick them off but tell them what happened in case they return to
		 * this terminal */
		(void)write_to_tty(ui->tty, IDLE_LOGOUT_MESSAGE, (unsigned long long)(allowed_idletime / 60ULL), idletime / 60ULL);
		syslog(LOG_INFO, "logging out user %s (idle for %llu, allowed %qd)", ui->pw->pw_name, idletime, allowed_idletime);
		/* do the actual log out */
		if (! LOGOUT(ui)) {
			syslog(LOG_WARNING, "failed to log out user %s on %s", ui->pw->pw_name, ui->tty);
		}
		/* thats it! */
		return;
	}

	/* if we reach here then the user is allowed to idle for now */
	if (debug) {
		syslog(LOG_DEBUG, "user %s allowed to stay (idle for %llu, allowed %qd)", ui->pw->pw_name, idletime, allowed_idletime);
	}

	/* user hasn't reached their limit but it may be time for a warning */
	for (i = 0; i < (sizeof(warn_times) / sizeof(int)); ++i) {
		/* convert time to seconds */
		warn_times[i] /= 60U;
		/* see if the warn time has passed */
		if ((allowed_idletime - idletime) <= warn_times[i]) {
			/* we are in the warning period but if we are well inside then we
			 * probably wanred the user last time around */
			if ((warn_times[i] - (allowed_idletime - idletime)) < monitor_interval) {
				/* only just inside so warn them */
				(void)write_to_tty(ui->tty, IDLE_WARN_MESSAGE, (unsigned long long)(allowed_idletime / 60ULL), idletime / 60ULL);

				if (debug) {
					syslog(LOG_DEBUG, "user %s given their %u minute warning", ui->pw->pw_name, warn_times[i] / 60U);
				}

				/* only one warning at a time - if warning times are very
				 * close together and monitor_interval is large then this
				 * can cause warnings to be skipped */
				break;
			} /* just inside warning period */
		} /* inside warning period */
	}

	return;
}
