#ifndef lint
static const char rcsid[] =
	"$Id: kill.c,v 1.4 2001/08/04 14:11:31 andrew Exp $";
#endif

#include <sys/types.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/stat.h>
#include <sys/sysctl.h>
#include <sys/user.h>
#include <sys/wait.h>

#include <fcntl.h>
#include <kvm.h>
#include <limits.h>
#include <paths.h>
#include <signal.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <termios.h>
#include <unistd.h>

#include "blimitd.h"
#include "config.h"
#include "kill.h"

/* Log the user out by setting the baudrate of their terminal to 0.
 *
 * Probably works OK on modem users but it doesn't kill off ssh/telnet users.
 */
int kill_hangup(struct user_instance *ui) {
	unsigned int success;
	int fd;
	struct termios tio;

	success = 1;

	/* get a fd on their terminal */
	if ((fd = open(ui->tty, O_RDWR, 0600)) == -1) {
		syslog(LOG_ERR, "open of %s to kill user %s failed: %m", ui->tty, ui->pw->pw_name);
		return(0);
	}

	/* wrap this so we can have one close handler at the bottom */
	do {
		/* set the baud rate to 0 */
		if (tcgetattr(fd, &tio) == -1) {
			syslog(LOG_ERR, "tcgetattr on %s for user %s failed: %m", ui->tty, ui->pw->pw_name);
			success = 0;
			break;
		}
		if (cfsetspeed(&tio, B0) == -1) {
			syslog(LOG_ERR, "cfsetspeed on %s for user %s failed: %m", ui->tty, ui->pw->pw_name);
			success = 0;
			break;
		}
		if (tcsetattr(fd, TCSADRAIN, &tio) == -1) {
			syslog(LOG_ERR, "tcsetattr on %s for user %s failed: %m", ui->tty, ui->pw->pw_name);
			success = 0;
			break;
		}
	} while(0);

	if (close(fd) == -1) {
		/* close failed */
		syslog(LOG_ERR, "close of %s to kill user %s failed: %m", ui->tty, ui->pw->pw_name);
	}

	return(success);
}

/* Log the user out by "revoking" their terminal.
 *
 * This causes the next read of the users terminal to return EOF.
 * This seems to work with ssh and telnet - haven't tried serial. Alas buggy
 * programs, such as old versions of PINE, have been known to go into infinite
 * loops when this happens.
 */
int kill_revoke(struct user_instance *ui) {

	if (revoke(ui->tty) == -1) {
		syslog(LOG_ERR, "revoke of %s to kill user %s failed: %m", ui->tty, ui->pw->pw_name);

		return(0);
	} else {
		return(1);
	}
}

/* Log the user out by sending a hangup signal to their shell.
 *
 * The shell is assumed to be the session leader of the session that the
 * first process we find running on the terminal is a member of. We find that
 * out by grovelling through kernel memory so this is probably completely
 * unportable and means blimitd must be recompiled if things happen in the
 * kernel (if you have to recompile ps the blimitd probably needs it as well).
 */
int kill_kvm(struct user_instance *ui) {
	char err_buffer[_POSIX2_LINE_MAX];
	kvm_t *kd;
	struct kinfo_proc *processes;
	struct proc session_leader;
	struct session tty_session;
	struct stat device_info;
	int number_of_processes,
		return_value;

	/* assume failure - its a good idea */
	return_value = 0;

	/* find the dev_t of the tty (used to identify the tty within the kernel
	 */
	if (stat(ui->tty, &device_info) != 0) {
		syslog(LOG_ERR, "stat of %s failed (%s:%d): %m", ui->tty, __FILE__, __LINE__);
		return(0);
	}

	/* open the kvm interface */
	if ((kd = kvm_openfiles(NULL, NULL, NULL, O_RDONLY, err_buffer)) == NULL) {
		syslog(LOG_ERR, "kvm_open failed (%s:%d): %s", __FILE__, __LINE__, err_buffer);
		return(0);
	}

	/* fetch all the processes running on the tty */
	if ((processes = kvm_getprocs(kd, KERN_PROC_TTY, (int)device_info.st_rdev, &number_of_processes)) == NULL) {
		syslog(LOG_ERR, "kvm_getprocs failed (%s:%d): %s", __FILE__, __LINE__, kvm_geterr(kd));
		goto failure;
	}

	/* we now have an array of "struct kinfo_proc"s (see sys/user.h) - one
	 * for each process we found on the tty. Each kinfo_proc has a pointer to
	 * the tty session in which the process participates. Alas the pointer is
	 * a kernel address (not an address in our address space) so we can't
	 * dereference it directly...we have to call kvm_read. We only need to
	 * look at the first process's kinfo_proc as all processes should be in
	 * the same session. The session struct (see sys/proc.h) has a pointer to
	 * a struct proc (see sys/proc.h) that represents the session leader. The
	 * wrong address space problem occurs here as well. This struct proc
	 * contains the pid of the session leader.
	 */

	 /* copy the session structure to our address space */
	if (kvm_read(kd, (unsigned long)processes[0].kp_eproc.e_tsess, &tty_session, sizeof(tty_session)) != sizeof(tty_session)) {
		syslog(LOG_ERR, "kvm_read failed (%s:%d): %s", __FILE__, __LINE__, kvm_geterr(kd));
		goto failure;
	}

	/* copy the session leader's struct proc to our address space */
	if (kvm_read(kd, (unsigned long)tty_session.s_leader, &session_leader, sizeof(session_leader)) != sizeof(session_leader)) {
		syslog(LOG_ERR, "kvm_read failed (%s:%d): %s", __FILE__, __LINE__, kvm_geterr(kd));
		goto failure;
	}

	/* send a hangup signal to the shell */
	if (kill(session_leader.p_pid, SIGHUP) != 0) {
		syslog(LOG_ERR, "kill failed (%s:%d): %m", __FILE__, __LINE__);
		goto failure;
	} else {
		/* success? well possibly..we don't actually check the process went */
		return_value = 1;
	}

	/* we skip to here if things fail so we always close the kvm interface.
	 * we could have used massive if staements or do/while(0) and break but
	 * we didn't */
failure:
	/* close the kvm interface */
	if (kvm_close(kd) != 0) {
		syslog(LOG_ERR, "kvm_close failed (%s:%d): %s", __FILE__, __LINE__, kvm_geterr(kd));
	}

	return(return_value);
}

/* Log the user out by running an external program.
 *
 * This is hopefully a fully portable method. If none of the above work then
 * this one should. You have to provide your own script though. The external
 * program is called with the arguments "-u <username> -t tty".
 *
 * NB. If a user can write arbitary data to their utmp record then using this
 * function will allow them to gain root access. If this is a possibility on
 * your system and none of the other kill_ functions work for you then it
 * isn't too hard to fix this (fork and exec).
 */
int kill_ext(struct user_instance *ui) {
	char *cmd_string;
	int sys_rv, rv;
	size_t cmd_length;

	rv = 0;

	/* determine the length of "cmd_name -u username -t tty" (dont forget to
	 * allow for the '\0') */
	cmd_length = strlen(KILL_CMD) + strlen(ui->pw->pw_name) + strlen(ui->tty) + 9;

	/* allocate memory for the string to be passed to system */
	if ((cmd_string = (char *)malloc(cmd_length)) == NULL) {
		syslog(LOG_ERR, "malloc failed: %m");
		return(0);
	}

	/* build the string */
	if (snprintf(cmd_string, cmd_length, "%s -u %s -t %s", KILL_CMD, ui->pw->pw_name, ui->tty) >= (int)cmd_length) {
		syslog(LOG_ERR, "cmd_string buffer too small in kill_ext (%s:%d)", __FILE__, __LINE__);
	} else {
		/* if we managed to build the string then we will try to run it */
		switch (sys_rv = system(cmd_string)) {
		case -1:
			/* system had problems */
			syslog(LOG_ERR, "system failed (%s:%d): %m", __FILE__, __LINE__);
			break;
		case 127:
			/* /bin/sh wouldn't run */
			syslog(LOG_ERR, "system failed (%s unavailable) (%s:%d): %m", _PATH_BSHELL, __FILE__, __LINE__);
			break;
		default:
			/* external program was run but it died with a signal */
			if (WIFSIGNALED(sys_rv)) {
				syslog(LOG_ERR, "%s exited on signal %d (%s:%d)", KILL_CMD, WTERMSIG(sys_rv), __FILE__, __LINE__);
			/* external program was run but it was stopped (suspended?) */
			} else if (WIFSTOPPED(sys_rv)) {
				syslog(LOG_ERR, "%s stopped on signal %d (%s:%d)", KILL_CMD, WSTOPSIG(sys_rv), __FILE__, __LINE__);
			/* program exited normally but with a non zero exit value */
			} else if (WIFEXITED(sys_rv) && (WEXITSTATUS(sys_rv) != 0)) {
				syslog(LOG_ERR, "%s exited with code %d (%s:%d)", KILL_CMD, WEXITSTATUS(sys_rv), __FILE__, __LINE__);
			} else {
				/* success */
				rv = 1;
			}
			break;
		}
	}

	/* free the string we passed to system */
	free(cmd_string);

	return(rv);
}
